import java.util.ArrayList;
import java.util.List;
import java.util.Collections;
import java.io.IOException;

import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.JDOMException;


/**
 * Class that allows for generating a model to be used with FeatureIDE based
 * on available gazebomodels and sensorconfigs.
 */
public class ModelSetup extends Setup {
	//Variables used exclusively by ModelSetup
	public final Document model;
	public final List<String> scAutoList;
	public final List<String> scOtherList;
	public final List<String> gazeboModelNamesList;
	
	/**
	 * Creates a runnable ModelSetup.
	 * 
	 * When reading input, scAutoPath, scOtherPath and in_gxPath may contain other 
	 * files that will be filtered out.
	 * 
	 * @throws IOException 
	 * @throws JDOMException 
	 */
	public ModelSetup() throws JDOMException, IOException {
		super();
		
		//Get the model template file as Document
		model = XMLHelper.readXml(in_mPath+in_mName);
		
		//Get all available sensorconfig file names from .../autogenerated/ directory
		scAutoList = new ArrayList<String>();
		
		List<String> scAutoDirectoryFilesList = XMLHelper.getFileNames(scAutoPath);
		for(String s : scAutoDirectoryFilesList) {
			if(s.endsWith(SC_ENDING)) {
				scAutoList.add(s);
			}
		}
		
		//Get all available sensorconfig file names from .../other/ directory
		scOtherList = new ArrayList<String>();
		
		List<String> scOtherDirectoryFilesList = XMLHelper.getFileNames(scOtherPath);
		for(String s : scOtherDirectoryFilesList) {
			if(s.endsWith(SC_ENDING)) {
				scOtherList.add(s);
			}
		}
		
		//Get all available gazeboModel file names
		gazeboModelNamesList = new ArrayList<String>();
		
		List<String> gxDirectoryFilesList = XMLHelper.getFileNames(in_gxPath);
		for(String s : gxDirectoryFilesList) {
			if(s.endsWith(GX_ENDING)) {
				gazeboModelNamesList.add(s);
			}
		}
	}
	
	/**
	 * Run the ModelSetup.
	 * 
	 * @throws IOException 
	 */
	public void run() throws IOException {
		System.out.println("ModelSetup started.");
		
		//Check for empty feature input
		this.checkIfEmpty();
		
		//Check whether there are no duplicate sensorconfig names
		this.checkForDistinctSensorConfigNames();
		
		//Sort features alphabetically
		this.sortFeatures();
		
		//Add features to model
		this.addFeatures();
		
		//Output model
		this.output();
		
		System.out.println("ModelSetup finished.");
	}
	
	/**
	 * Check whether feature input was empty. 
	 * It is possible to have no sensors (so you can build a custom bot with no sensors), 
	 * but at least one bot template is required!
	 */
	private void checkIfEmpty() {
		if(this.scAutoList.isEmpty()) {
			System.out.println("Warning: No sensorconfig \'"+SC_ENDING+"\' files found in \'"+this.scAutoPath+"\'.");
		}
		if(this.scOtherList.isEmpty()) {
			System.out.println("Warning: No sensorconfig \'"+SC_ENDING+"\' files found in \'"+this.scOtherPath+"\'.");
		}
		if(this.gazeboModelNamesList.isEmpty()) {
			System.err.println("Error: No gazebo model \'"+GX_ENDING+"\' files found in \'"+this.in_gxPath+"\'.");
			System.exit(-1);
		}
	}
	
	/**
	 * Checks whether sensorconfigs in .../autogenerated/ and .../other/ have distinct names. 
	 * Thid could happen because we read from two directories. 
	 * FeatureIDE does not allow for duplicate feature names. 
	 */
	private void checkForDistinctSensorConfigNames() {
		List<String> testList = new ArrayList<String>();
		testList.addAll(this.scAutoList);
		testList.addAll(this.scOtherList);
		Collections.sort(testList);
		
		if(!testList.isEmpty()) {
			for(int i = 0; i < testList.size()-1; i++) {
				String name1 = testList.get(i);
				String name2 = testList.get(i+1);
				if(name1.equals(name2)) {
					System.err.println("Error: At least two sensorconfig files have the same name: \'"+name1+"\'");
					System.exit(-1);
				}
			}
		}
	}
	
	/**
	 * Sorts scAutoList, scOtherList, gazeboModelNamesList alphabetically. 
	 * (ascending order; upper case first; then lower case)
	 */
	private void sortFeatures() {
		Collections.sort(this.scAutoList);
		Collections.sort(this.scOtherList);
		Collections.sort(this.gazeboModelNamesList);
	}
	
	/**
	 * Adds all features in gazeboModelNamesList, scOtherList, scAutoList to model.
	 */
	private void addFeatures() {
		//Get Elements in model where features are to be added
		List<Element> templatesList = XMLHelper.xQueryElements(this.model, "//*[@name='Templates']");
		if(templatesList.size()!=1) {
			System.err.println("Error when parsing for \'Templates\' in \'"+this.in_mName+"\'.");
			System.exit(-1);
		}
		Element templates = templatesList.get(0);
		
		List<Element> customList = XMLHelper.xQueryElements(this.model, "//*[@name='Custom']");
		if(customList.size()!=1) {
			System.err.println("Error when parsing for \'Custom\' in \'"+this.in_mName+"\'.");
			System.exit(-1);
		}
		Element custom = customList.get(0);
		
		//Add features to model
		for(String gm : this.gazeboModelNamesList) {
			String featureName = gm.substring(0, gm.length() - GX_ENDING.length());
			Element feature = new Element("feature").setAttribute("name", featureName);
			templates.addContent(feature);
		}
		
		for(String sc : this.scOtherList) {
			String featureName = sc.substring(0, sc.length() - SC_ENDING.length());
			Element feature = new Element("feature").setAttribute("name", featureName);
			custom.addContent(feature);
		}
		
		for(String sc : this.scAutoList) {
			String featureName = sc.substring(0, sc.length() - SC_ENDING.length());
			Element feature = new Element("feature").setAttribute("name", featureName);
			custom.addContent(feature);
		}
	}
	
	/**
	 * Output model as XML file to be used by FeatureIDE.
	 * 
	 * @throws IOException 
	 */
	private void output() throws IOException {
		XMLHelper.saveAsXml(this.model, this.out_mPath+this.out_mName);
	}
}
